@@ -13,9 +13,9 @@
 >>> doc_id, doc_rev = db.save({'type': 'Person', 'name': 'John Doe'})
 >>> doc = db[doc_id]
 >>> doc['type']
-'Person'
+u'Person'
 >>> doc['name']
-'John Doe'
+u'John Doe'
 >>> del db[doc.id]
 >>> doc.id in db
 False
@@ -23,6 +23,7 @@ False
 >>> del server['python-tests']
 """
 
+import itertools
 import mimetypes
 import os
 from types import FunctionType
@@ -31,7 +32,7 @@ from textwrap import dedent
 import re
 import warnings
 
-from couchdb import http, json
+from couchdb import http, json, util
 
 __all__ = ['Server', 'Database', 'Document', 'ViewResults', 'Row']
 __docformat__ = 'restructuredtext en'
@@ -75,7 +76,7 @@ class Server(object):
         :param full_commit: turn on the X-Couch-Full-Commit header
         :param session: an http.Session instance or None for a default session
         """
-        if isinstance(url, basestring):
+        if isinstance(url, util.strbase):
             self.resource = http.Resource(url, session or http.Session())
         else:
             self.resource = url # treat as a Resource object
@@ -90,7 +91,7 @@ class Server(object):
         :return: `True` if a database with the name exists, `False` otherwise
         """
         try:
-            self.resource.head(validate_dbname(name))
+            self.resource.head(name)
             return True
         except http.ResourceNotFound:
             return False
@@ -113,6 +114,9 @@ class Server(object):
         except:
             return False
 
+    def __bool__(self):
+        return self.__nonzero__()
+
     def __repr__(self):
         return '<%s %r>' % (type(self).__name__, self.resource.url)
 
@@ -122,7 +126,7 @@ class Server(object):
         :param name: the name of the database
         :raise ResourceNotFound: if no database with that name exists
         """
-        self.resource.delete_json(validate_dbname(name))
+        self.resource.delete_json(name)
 
     def __getitem__(self, name):
         """Return a `Database` object representing the database with the
@@ -133,7 +137,7 @@ class Server(object):
         :rtype: `Database`
         :raise ResourceNotFound: if no database with that name exists
         """
-        db = Database(self.resource(name), validate_dbname(name))
+        db = Database(self.resource(name), name)
         db.resource.head() # actually make a request to the database
         return db
 
@@ -159,9 +163,17 @@ class Server(object):
         status, headers, data = self.resource.get_json()
         return data['version']
 
-    def stats(self):
-        """Database statistics."""
-        status, headers, data = self.resource.get_json('_stats')
+    def stats(self, name=None):
+        """Server statistics.
+
+        :param name: name of single statistic, e.g. httpd/requests
+                     (None -- return all statistics)
+        """
+        if not name:
+            resource = self.resource('_stats')
+        else:
+            resource = self.resource('_stats', *name.split('/'))
+        status, headers, data = resource.get_json()
         return data
 
     def tasks(self):
@@ -190,7 +202,7 @@ class Server(object):
         :rtype: `Database`
         :raise PreconditionFailed: if a database with that name already exists
         """
-        self.resource.put_json(validate_dbname(name))
+        self.resource.put_json(name)
         return self[name]
 
     def delete(self, name):
@@ -230,18 +242,18 @@ class Database(object):
 
     >>> doc = db[doc_id]
     >>> doc                 #doctest: +ELLIPSIS
-    <Document '...'@... {...}>
+    <Document u'...'@... {...}>
 
     Documents are represented as instances of the `Row` class, which is
     basically just a normal dictionary with the additional attributes ``id`` and
     ``rev``:
 
     >>> doc.id, doc.rev     #doctest: +ELLIPSIS
-    ('...', ...)
+    (u'...', ...)
     >>> doc['type']
-    'Person'
+    u'Person'
     >>> doc['name']
-    'John Doe'
+    u'John Doe'
 
     To update an existing document, you use item access, too:
 
@@ -263,7 +275,7 @@ class Database(object):
     """
 
     def __init__(self, url, name=None, session=None):
-        if isinstance(url, basestring):
+        if isinstance(url, util.strbase):
             if not url.startswith('http'):
                 url = DEFAULT_BASE_URL + url
             self.resource = http.Resource(url, session)
@@ -282,7 +294,7 @@ class Database(object):
         :return: `True` if a document with the ID exists, `False` otherwise
         """
         try:
-            self.resource.head(id)
+            _doc_resource(self.resource, id).head()
             return True
         except http.ResourceNotFound:
             return False
@@ -304,13 +316,17 @@ class Database(object):
         except:
             return False
 
+    def __bool__(self):
+        return self.__nonzero__()
+
     def __delitem__(self, id):
         """Remove the document with the specified ID from the database.
 
         :param id: the document ID
         """
-        status, headers, data = self.resource.head(id)
-        self.resource.delete_json(id, rev=headers['etag'].strip('"'))
+        resource = _doc_resource(self.resource, id)
+        status, headers, data = resource.head()
+        resource.delete_json(rev=headers['etag'].strip('"'))
 
     def __getitem__(self, id):
         """Return the document with the specified ID.
@@ -319,7 +335,7 @@ class Database(object):
         :return: a `Row` object representing the requested document
         :rtype: `Document`
         """
-        _, _, data = self.resource.get_json(id)
+        _, _, data = _doc_resource(self.resource, id).get_json()
         return Document(data)
 
     def __setitem__(self, id, content):
@@ -330,7 +346,8 @@ class Database(object):
                         new documents, or a `Row` object for existing
                         documents
         """
-        status, headers, data = self.resource.put_json(id, body=content)
+        resource = _doc_resource(self.resource, id)
+        status, headers, data = resource.put_json(body=content)
         content.update({'_id': data['id'], '_rev': data['rev']})
 
     @property
@@ -346,6 +363,14 @@ class Database(object):
             self.info()
         return self._name
 
+    @property
+    def security(self):
+        return self.resource.get_json('_security')[2]
+
+    @security.setter
+    def security(self, doc):
+        self.resource.put_json('_security', body=doc)
+
     def create(self, data):
         """Create a new document in the database with a random ID that is
         generated by the server.
@@ -401,7 +426,7 @@ class Database(object):
         :rtype: `tuple`
         """
         if '_id' in doc:
-            func = self.resource(doc['_id']).put_json
+            func = _doc_resource(self.resource, doc['_id']).put_json
         else:
             func = self.resource.post_json
         _, _, data = func(body=doc, **options)
@@ -411,6 +436,18 @@ class Database(object):
             doc['_rev'] = rev
         return id, rev
 
+    def cleanup(self):
+        """Clean up old design document indexes.
+
+        Remove all unused index files from the database storage area.
+
+        :return: a boolean to indicate successful cleanup initiation
+        :rtype: `bool`
+        """
+        headers = {'Content-Type': 'application/json'}
+        _, _, data = self.resource('_view_cleanup').post_json(headers=headers)
+        return data['ok']
+
     def commit(self):
         """If the server is configured to delay commits, or previous requests
         used the special ``X-Couch-Full-Commit: false`` header to disable
@@ -453,7 +490,7 @@ class Database(object):
         :rtype: `str`
         :since: 0.6
         """
-        if not isinstance(src, basestring):
+        if not isinstance(src, util.strbase):
             if not isinstance(src, dict):
                 if hasattr(src, 'items'):
                     src = dict(src.items())
@@ -462,7 +499,7 @@ class Database(object):
                                     type(src))
             src = src['_id']
 
-        if not isinstance(dest, basestring):
+        if not isinstance(dest, util.strbase):
             if not isinstance(dest, dict):
                 if hasattr(dest, 'items'):
                     dest = dict(dest.items())
@@ -477,7 +514,7 @@ class Database(object):
 
         _, _, data = self.resource._request('COPY', src,
                                             headers={'Destination': dest})
-        data = json.decode(data.read())
+        data = json.decode(data.read().decode('utf-8'))
         return data['rev']
 
     def delete(self, doc):
@@ -496,10 +533,10 @@ class Database(object):
         >>> doc2 = db['johndoe']
         >>> doc2['age'] = 42
         >>> db['johndoe'] = doc2
-        >>> db.delete(doc)
+        >>> db.delete(doc) # doctest: +IGNORE_EXCEPTION_DETAIL
         Traceback (most recent call last):
           ...
-        ResourceConflict: ('conflict', 'Document update conflict.')
+        ResourceConflict: (u'conflict', u'Document update conflict.')
 
         >>> del server['python-tests']
 
@@ -509,7 +546,7 @@ class Database(object):
         """
         if doc['_id'] is None:
             raise ValueError('document ID cannot be None')
-        self.resource.delete_json(doc['_id'], rev=doc['_rev'])
+        _doc_resource(self.resource, doc['_id']).delete_json(rev=doc['_rev'])
 
     def get(self, id, default=None, **options):
         """Return the document with the specified ID.
@@ -522,7 +559,7 @@ class Database(object):
         :rtype: `Document`
         """
         try:
-            _, _, data = self.resource.get_json(id, **options)
+            _, _, data = _doc_resource(self.resource, id).get_json(**options)
         except http.ResourceNotFound:
             return default
         if hasattr(data, 'items'):
@@ -538,7 +575,8 @@ class Database(object):
                  in reverse chronological order, if any were found
         """
         try:
-            status, headers, data = self.resource.get_json(id, revs=True)
+            resource = _doc_resource(self.resource, id)
+            status, headers, data = resource.get_json(revs=True)
         except http.ResourceNotFound:
             return
 
@@ -550,18 +588,25 @@ class Database(object):
                 return
             yield revision
 
-    def info(self):
-        """Return information about the database as a dictionary.
+    def info(self, ddoc=None):
+        """Return information about the database or design document as a
+        dictionary.
+
+        Without an argument, returns database information. With an argument,
+        return information for the given design document.
 
         The returned dictionary exactly corresponds to the JSON response to
-        a ``GET`` request on the database URI.
+        a ``GET`` request on the database or design document's info URI.
 
         :return: a dictionary of database properties
         :rtype: ``dict``
         :since: 0.4
         """
-        _, _, data = self.resource.get_json()
-        self._name = data['db_name']
+        if ddoc is not None:
+            _, _, data = self.resource('_design', ddoc, '_info').get_json()
+        else:
+            _, _, data = self.resource.get_json()
+            self._name = data['db_name']
         return data
 
     def delete_attachment(self, doc, filename):
@@ -576,7 +621,7 @@ class Database(object):
         :param filename: the name of the attachment file
         :since: 0.4.1
         """
-        resource = self.resource(doc['_id'])
+        resource = _doc_resource(self.resource, doc['_id'])
         _, _, data = resource.delete_json(filename, rev=doc['_rev'])
         doc['_rev'] = data['rev']
 
@@ -593,12 +638,12 @@ class Database(object):
                  of the `default` argument if the attachment is not found
         :since: 0.4.1
         """
-        if isinstance(id_or_doc, basestring):
+        if isinstance(id_or_doc, util.strbase):
             id = id_or_doc
         else:
             id = id_or_doc['_id']
         try:
-            _, _, data = self.resource(id).get(filename)
+            _, _, data = _doc_resource(self.resource, id).get(filename)
             return data
         except http.ResourceNotFound:
             return default
@@ -632,7 +677,7 @@ class Database(object):
                 filter(None, mimetypes.guess_type(filename))
             )
 
-        resource = self.resource(doc['_id'])
+        resource = _doc_resource(self.resource, doc['_id'])
         status, headers, data = resource.put_json(filename, body=content, headers={
             'Content-Type': content_type
         }, rev=doc['_rev'])
@@ -652,17 +697,17 @@ class Database(object):
         ...         emit(doc.name, null);
         ... }'''
         >>> for row in db.query(map_fun):
-        ...     print row.key
+        ...     print(row.key)
         John Doe
         Mary Jane
 
         >>> for row in db.query(map_fun, descending=True):
-        ...     print row.key
+        ...     print(row.key)
         Mary Jane
         John Doe
 
         >>> for row in db.query(map_fun, key='John Doe'):
-        ...     print row.key
+        ...     print(row.key)
         John Doe
 
         >>> del server['python-tests']
@@ -692,10 +737,10 @@ class Database(object):
         ...     Document(type='Person', name='Mary Jane'),
         ...     Document(type='City', name='Gotham City')
         ... ]):
-        ...     print repr(doc) #doctest: +ELLIPSIS
-        (True, '...', '...')
-        (True, '...', '...')
-        (True, '...', '...')
+        ...     print(repr(doc)) #doctest: +ELLIPSIS
+        (True, u'...', u'...')
+        (True, u'...', u'...')
+        (True, u'...', u'...')
 
         >>> del server['python-tests']
 
@@ -750,6 +795,25 @@ class Database(object):
 
         return results
 
+    def purge(self, docs):
+        """Perform purging (complete removing) of the given documents.
+
+        Uses a single HTTP request to purge all given documents. Purged
+        documents do not leave any meta-data in the storage and are not
+        replicated.
+        """
+        content = {}
+        for doc in docs:
+            if isinstance(doc, dict):
+                content[doc['_id']] = [doc['_rev']]
+            elif hasattr(doc, 'items'):
+                doc = dict(doc.items())
+                content[doc['_id']] = [doc['_rev']]
+            else:
+                raise TypeError('expected dict, got %s' % type(doc))
+        _, _, data = self.resource.post_json('_purge', body=content)
+        return data
+
     def view(self, name, wrapper=None, **options):
         """Execute a predefined view.
 
@@ -758,7 +822,7 @@ class Database(object):
         >>> db['gotham'] = dict(type='City', name='Gotham City')
 
         >>> for row in db.view('_all_docs'):
-        ...     print row.id
+        ...     print(row.id)
         gotham
 
         >>> del server['python-tests']
@@ -773,19 +837,112 @@ class Database(object):
         :return: the view results
         :rtype: `ViewResults`
         """
-        if not name.startswith('_'):
-            design, name = name.split('/', 1)
-            name = '/'.join(['_design', design, '_view', name])
-        return PermanentView(self.resource(*name.split('/')), name,
+        path = _path_from_name(name, '_view')
+        return PermanentView(self.resource(*path), '/'.join(path),
                              wrapper=wrapper)(**options)
 
+    def iterview(self, name, batch, wrapper=None, **options):
+        """Iterate the rows in a view, fetching rows in batches and yielding
+        one row at a time.
+
+        Since the view's rows are fetched in batches any rows emitted for
+        documents added, changed or deleted between requests may be missed or
+        repeated.
+
+        :param name: the name of the view; for custom views, use the format
+                     ``design_docid/viewname``, that is, the document ID of the
+                     design document and the name of the view, separated by a
+                     slash.
+        :param batch: number of rows to fetch per HTTP request.
+        :param wrapper: an optional callable that should be used to wrap the
+                        result rows
+        :param options: optional query string parameters
+        :return: row generator
+        """
+        # Check sane batch size.
+        if batch <= 0:
+            raise ValueError('batch must be 1 or more')
+        # Save caller's limit, it must be handled manually.
+        limit = options.get('limit')
+        if limit is not None and limit <= 0:
+            raise ValueError('limit must be 1 or more')
+        while True:
+            loop_limit = min(limit or batch, batch)
+            # Get rows in batches, with one extra for start of next batch.
+            options['limit'] = loop_limit + 1
+            rows = list(self.view(name, wrapper, **options))
+            # Yield rows from this batch.
+            for row in itertools.islice(rows, loop_limit):
+                yield row
+            # Decrement limit counter.
+            if limit is not None:
+                limit -= min(len(rows), batch)
+            # Check if there is nothing else to yield.
+            if len(rows) <= batch or (limit is not None and limit == 0):
+                break
+            # Update options with start keys for next loop.
+            options.update(startkey=rows[-1]['key'], startkey_docid=rows[-1]['id'])
+
+    def show(self, name, docid=None, **options):
+        """Call a 'show' function.
+
+        :param name: the name of the show function in the format
+                     ``designdoc/showname``
+        :param docid: optional ID of a document to pass to the show function.
+        :param options: optional query string parameters
+        :return: (headers, body) tuple, where headers is a dict of headers
+                 returned from the show function and body is a readable
+                 file-like instance
+        """
+        path = _path_from_name(name, '_show')
+        if docid:
+            path.append(docid)
+        status, headers, body = self.resource(*path).get(**options)
+        return headers, body
+
+    def list(self, name, view, **options):
+        """Format a view using a 'list' function.
+
+        :param name: the name of the list function in the format
+                     ``designdoc/listname``
+        :param view: the name of the view in the format ``designdoc/viewname``
+        :param options: optional query string parameters
+        :return: (headers, body) tuple, where headers is a dict of headers
+                 returned from the list function and body is a readable
+                 file-like instance
+        """
+        path = _path_from_name(name, '_list')
+        path.extend(view.split('/', 1))
+        _, headers, body = _call_viewlike(self.resource(*path), options)
+        return headers, body
+
+    def update_doc(self, name, docid=None, **options):
+        """Calls server side update handler.
+
+        :param name: the name of the update handler function in the format
+                     ``designdoc/updatename``.
+        :param docid: optional ID of a document to pass to the update handler.
+        :param options: optional query string parameters.
+        :return: (headers, body) tuple, where headers is a dict of headers
+                 returned from the list function and body is a readable
+                 file-like instance
+        """
+        path = _path_from_name(name, '_update')
+        if docid is None:
+            func = self.resource(*path).post
+        else:
+            path.append(docid)
+            func = self.resource(*path).put
+        _, headers, body = func(**options)
+        return headers, body
+
     def _changes(self, **opts):
         _, _, data = self.resource.get('_changes', **opts)
-        lines = iter(data)
+        lines = data.iterchunks()
         for ln in lines:
             if not ln: # skip heartbeats
                 continue
-            doc = json.decode(ln)
+            doc = json.decode(ln.decode('utf-8'))
             if 'last_seq' in doc: # consume the rest of the response if this
                 for ln in lines:  # was the last line, allows conn reuse
                     pass
@@ -794,7 +951,8 @@ class Database(object):
     def changes(self, **opts):
         """Retrieve a changes feed from the database.
 
-        Takes since, feed, heartbeat and timeout options.
+        :param opts: optional query string parameters
+        :return: an iterable over change notification dicts
         """
         if opts.get('feed') == 'continuous':
             return self._changes(**opts)
@@ -802,6 +960,26 @@ class Database(object):
         return data
 
 
+def _doc_resource(base, doc_id):
+    """Return the resource for the given document id.
+    """
+    # Split an id that starts with a reserved segment, e.g. _design/foo, so
+    # that the / that follows the 1st segment does not get escaped.
+    if doc_id[:1] == '_':
+        return base(*doc_id.split('/', 1))
+    return base(doc_id)
+
+
+def _path_from_name(name, type):
+    """Expand a 'design/foo' style name to its full path as a list of
+    segments.
+    """
+    if name.startswith('_'):
+        return name.split('/')
+    design, name = name.split('/', 1)
+    return ['_design', design, type, name]
+
+
 class Document(dict):
     """Representation of a document in the database.
 
@@ -835,7 +1013,7 @@ class View(object):
     """Abstract representation of a view or query."""
 
     def __init__(self, url, wrapper=None, session=None):
-        if isinstance(url, basestring):
+        if isinstance(url, util.strbase):
             self.resource = http.Resource(url, session)
         else:
             self.resource = url
@@ -847,15 +1025,6 @@ class View(object):
     def __iter__(self):
         return iter(self())
 
-    def _encode_options(self, options):
-        retval = {}
-        for name, value in options.items():
-            if name in ('key', 'startkey', 'endkey') \
-                    or not isinstance(value, basestring):
-                value = json.encode(value)
-            retval[name] = value
-        return retval
-
     def _exec(self, options):
         raise NotImplementedError
 
@@ -871,13 +1040,7 @@ class PermanentView(View):
         return '<%s %r>' % (type(self).__name__, self.name)
 
     def _exec(self, options):
-        if 'keys' in options:
-            options = options.copy()
-            keys = {'keys': options.pop('keys')}
-            _, _, data = self.resource.post_json(body=keys,
-                                                 **self._encode_options(options))
-        else:
-            _, _, data = self.resource.get_json(**self._encode_options(options))
+        _, _, data = _call_viewlike(self.resource, options)
         return data
 
 
@@ -911,10 +1074,34 @@ class TemporaryView(View):
         content = json.encode(body).encode('utf-8')
         _, _, data = self.resource.post_json(body=content, headers={
             'Content-Type': 'application/json'
-        }, **self._encode_options(options))
+        }, **_encode_view_options(options))
         return data
 
 
+def _encode_view_options(options):
+    """Encode any items in the options dict that are sent as a JSON string to a
+    view/list function.
+    """
+    retval = {}
+    for name, value in options.items():
+        if name in ('key', 'startkey', 'endkey') \
+                or not isinstance(value, util.strbase):
+            value = json.encode(value)
+        retval[name] = value
+    return retval
+
+
+def _call_viewlike(resource, options):
+    """Call a resource that takes view-like options.
+    """
+    if 'keys' in options:
+        options = options.copy()
+        keys = {'keys': options.pop('keys')}
+        return resource.post_json(body=keys, **_encode_view_options(options))
+    else:
+        return resource.get_json(**_encode_view_options(options))
+
+
 class ViewResults(object):
     """Representation of a parameterized view (either permanent or temporary)
     and the results it produces.
@@ -944,7 +1131,7 @@ class ViewResults(object):
 
     >>> people = results[['Person']:['Person','ZZZZ']]
     >>> for person in people:
-    ...     print person.value
+    ...     print(person.value)
     John Doe
     Mary Jane
     >>> people.total_rows, people.offset
@@ -955,7 +1142,7 @@ class ViewResults(object):
     can still return multiple rows:
 
     >>> list(results[['City', 'Gotham City']])
-    [<Row id='gotham', key=['City', 'Gotham City'], value='Gotham City'>]
+    [<Row id=u'gotham', key=[u'City', u'Gotham City'], value=u'Gotham City'>]
 
     >>> del server['python-tests']
     """
@@ -963,7 +1150,7 @@ class ViewResults(object):
     def __init__(self, view, options):
         self.view = view
         self.options = options
-        self._rows = self._total_rows = self._offset = None
+        self._rows = self._total_rows = self._offset = self._update_seq = None
 
     def __repr__(self):
         return '<%s %r %r>' % (type(self).__name__, self.view, self.options)
@@ -992,6 +1179,7 @@ class ViewResults(object):
         self._rows = [wrapper(row) for row in data['rows']]
         self._total_rows = data.get('total_rows')
         self._offset = data.get('offset', 0)
+        self._update_seq = data.get('update_seq')
 
     @property
     def rows(self):
@@ -1027,16 +1215,28 @@ class ViewResults(object):
             self._fetch()
         return self._offset
 
+    @property
+    def update_seq(self):
+        """The database update sequence that the view reflects.
+
+        The update sequence is included in the view result only when it is
+        explicitly requested using the `update_seq=true` query option.
+        Otherwise, the value is None.
+
+        :rtype: `int` or `NoneType` depending on the query options
+        """
+        if self._rows is None:
+            self._fetch()
+        return self._update_seq
+
 
 class Row(dict):
     """Representation of a row as returned by database views."""
 
     def __repr__(self):
-        if self.id is None:
-            return '<%s key=%r, value=%r>' % (type(self).__name__, self.key,
-                                              self.value)
-        return '<%s id=%r, key=%r, value=%r>' % (type(self).__name__, self.id,
-                                                 self.key, self.value)
+        keys = 'id', 'key', 'error', 'value'
+        items = ['%s=%r' % (k, self[k]) for k in keys if k in self]
+        return '<%s %s>' % (type(self).__name__, ', '.join(items))
 
     @property
     def id(self):
@@ -1047,13 +1247,15 @@ class Row(dict):
 
     @property
     def key(self):
-        """The associated key."""
         return self['key']
 
     @property
     def value(self):
-        """The associated value."""
-        return self['value']
+        return self.get('value')
+
+    @property
+    def error(self):
+        return self.get('error')
 
     @property
     def doc(self):
@@ -1064,13 +1266,3 @@ class Row(dict):
         doc = self.get('doc')
         if doc:
             return Document(doc)
-
-
-SPECIAL_DB_NAMES = set(['_users'])
-VALID_DB_NAME = re.compile(r'^[a-z][a-z0-9_$()+-/]*$')
-def validate_dbname(name):
-    if name in SPECIAL_DB_NAMES:
-        return name
-    if not VALID_DB_NAME.match(name):
-        raise ValueError('Invalid database name')
-    return name
