@@ -15,7 +15,8 @@
 To define a document mapping, you declare a Python class inherited from
 `Document`, and add any number of `Field` attributes:
 
->>> from couchdb.mapping import TextField, IntegerField, DateField
+>>> from datetime import datetime
+>>> from couchdb.mapping import Document, TextField, IntegerField, DateTimeField
 >>> class Person(Document):
 ...     name = TextField()
 ...     age = IntegerField()
@@ -57,7 +58,7 @@ True
 >>> del server['python-tests']
 """
 
-import copy
+import copy, sys
 
 from calendar import timegm
 from datetime import date, datetime, time
@@ -65,6 +66,7 @@ from decimal import Decimal
 from time import strptime, struct_time
 
 from couchdb.design import ViewDefinition
+from couchdb import util
 
 __all__ = ['Mapping', 'Document', 'Field', 'TextField', 'FloatField',
            'IntegerField', 'LongField', 'BooleanField', 'DecimalField',
@@ -105,7 +107,7 @@ class Field(object):
         instance._data[self.name] = value
 
     def _to_python(self, value):
-        return unicode(value)
+        return util.utype(value)
 
     def _to_json(self, value):
         return self._to_python(value)
@@ -126,9 +128,10 @@ class MappingMeta(type):
         d['_fields'] = fields
         return type.__new__(cls, name, bases, d)
 
+MappingMetaClass = MappingMeta('MappingMetaClass', (object,), {})
 
-class Mapping(object):
-    __metaclass__ = MappingMeta
+
+class Mapping(MappingMetaClass):
 
     def __init__(self, **values):
         self._data = {}
@@ -153,7 +156,7 @@ class Mapping(object):
     def __setitem__(self, name, value):
         self._data[name] = value
 
-    def get(self, name, default):
+    def get(self, name, default=None):
         return self._data.get(name, default)
 
     def setdefault(self, name, default):
@@ -199,7 +202,7 @@ class ViewField(object):
     >>> Person.by_name
     <ViewDefinition '_design/people/_view/by_name'>
     
-    >>> print Person.by_name.map_fun
+    >>> print(Person.by_name.map_fun)
     function(doc) {
         emit(doc.name, doc);
     }
@@ -234,7 +237,7 @@ class ViewField(object):
     >>> Person.by_name
     <ViewDefinition '_design/people/_view/by_name'>
 
-    >>> print Person.by_name.map_fun
+    >>> print(Person.by_name.map_fun)
     def by_name(doc):
         yield doc['name'], doc
     """
@@ -291,9 +294,10 @@ class DocumentMeta(MappingMeta):
                     attrval.name = attrname
         return MappingMeta.__new__(cls, name, bases, d)
 
+DocumentMetaClass = DocumentMeta('DocumentMetaClass', (object,), {})
 
-class Document(Mapping):
-    __metaclass__ = DocumentMeta
+
+class Document(DocumentMetaClass, Mapping):
 
     def __init__(self, id=None, **values):
         Mapping.__init__(self, **values)
@@ -407,7 +411,7 @@ class Document(Mapping):
 
 class TextField(Field):
     """Mapping field for string values."""
-    _to_python = unicode
+    _to_python = util.utype
 
 
 class FloatField(Field):
@@ -422,7 +426,7 @@ class IntegerField(Field):
 
 class LongField(Field):
     """Mapping field for long integer values."""
-    _to_python = long
+    _to_python = util.ltype
 
 
 class BooleanField(Field):
@@ -437,7 +441,7 @@ class DecimalField(Field):
         return Decimal(value)
 
     def _to_json(self, value):
-        return unicode(value)
+        return util.utype(value)
 
 
 class DateField(Field):
@@ -453,7 +457,7 @@ class DateField(Field):
     """
 
     def _to_python(self, value):
-        if isinstance(value, basestring):
+        if isinstance(value, util.strbase):
             try:
                 value = date(*strptime(value, '%Y-%m-%d')[:3])
             except ValueError:
@@ -479,7 +483,7 @@ class DateTimeField(Field):
     """
 
     def _to_python(self, value):
-        if isinstance(value, basestring):
+        if isinstance(value, util.strbase):
             try:
                 value = value.split('.', 1)[0] # strip out microseconds
                 value = value.rstrip('Z') # remove timezone separator
@@ -509,7 +513,7 @@ class TimeField(Field):
     """
 
     def _to_python(self, value):
-        if isinstance(value, basestring):
+        if isinstance(value, util.strbase):
             try:
                 value = value.split('.', 1)[0] # strip out microseconds
                 value = time(*strptime(value, '%H:%M:%S')[3:6])
@@ -553,7 +557,7 @@ class DictField(Field):
     >>> post.author.email
     u'john@doe.com'
     >>> post.extra
-    {'foo': 'bar'}
+    {u'foo': u'bar'}
 
     >>> del server['python-tests']
     """
@@ -603,11 +607,11 @@ class ListField(Field):
     >>> post = Post.load(db, post.id)
     >>> comment = post.comments[0]
     >>> comment['author']
-    'myself'
+    u'myself'
     >>> comment['content']
-    'Bla bla'
+    u'Bla bla'
     >>> comment['time'] #doctest: +ELLIPSIS
-    '...T...Z'
+    u'...T...Z'
 
     >>> del server['python-tests']
     """
@@ -660,16 +664,24 @@ class ListField(Field):
             return str(self.list)
 
         def __unicode__(self):
-            return unicode(self.list)
+            return util.utype(self.list)
 
         def __delitem__(self, index):
-            del self.list[index]
+            if isinstance(index, slice):
+                self.__delslice__(index.start, index.stop)
+            else:
+                del self.list[index]
 
         def __getitem__(self, index):
+            if isinstance(index, slice):
+                return self.__getslice__(index.start, index.stop)
             return self.field._to_python(self.list[index])
 
         def __setitem__(self, index, value):
-            self.list[index] = self.field._to_json(value)
+            if isinstance(index, slice):
+                self.__setslice__(index.start, index.stop, value)
+            else:
+                self.list[index] = self.field._to_json(value)
 
         def __delslice__(self, i, j):
             del self.list[i:j]
