@@ -6,10 +6,23 @@
 # This software is licensed as described in the file COPYING, which
 # you should have received as part of this distribution.
 
+import doctest
 import random
+import re
 import sys
+
 from couchdb import client
 
+class Py23DocChecker(doctest.OutputChecker):
+    def check_output(self, want, got, optionflags):
+        if sys.version_info[0] > 2:
+            want = re.sub("u'(.*?)'", "'\\1'", want)
+            want = re.sub('u"(.*?)"', '"\\1"', want)
+        return doctest.OutputChecker.check_output(self, want, got, optionflags)
+
+def doctest_suite(mod):
+    return doctest.DocTestSuite(mod, checker=Py23DocChecker())
+
 class TempDatabaseMixin(object):
 
     temp_dbs = None
@@ -28,10 +41,9 @@ class TempDatabaseMixin(object):
             self.temp_dbs = {}
         # Find an unused database name
         while True:
-            name = 'couchdb-python/%d' % random.randint(0, sys.maxint)
+            name = 'couchdb-python/%d' % random.randint(0, sys.maxsize)
             if name not in self.temp_dbs:
                 break
-            print '%s already used' % name
         db = self.server.create(name)
         self.temp_dbs[name] = db
         return name, db
